% Konstantina Palla <kp376@cam.ac.uk>
% University of Cambridge
% Created: February 2012
% Modified: August 2012

function [param settings] = inference(param, settings)
% function that runs the inference process.
% input:    param - structure that contains the set of parameters
%           settings - structure that contains the experimental settings

%start the timer
tic

% check the settings are correctly set
[param settings] = check_settings(param, settings);

% initialise the model; initialise the structures
% person; variable that contains a random arrangement of the N objects
[param settings person] = initModel(param, settings);
N=size(param.R,1);


test_mask=settings.test_mask;
train_mask=settings.train_mask;

%% -------------------- Run the sampler ------------------------------- %
%print the settings of the current experiment
fprintf('Running sampler with \n sampler for z: %s \n sampler for c: %s\n sampler for w: %s\n is sequential-sampling as initialization on?: %s\n', ...
    settings.sampler_z, settings.sampler_c,  settings.sampler_w, settings.z_init);

np=0;
vec=[];

% Run sequential initialisation if it is chosen for Z
if (isequal(settings.z_init,'sequential') & (isequal(settings.sampler_z, 'metropolis_hastings') | isequal(settings.sampler_z, 'finite')))
    fprintf('number of total inner iterations for the sequential initialization: %d\n',settings.in_itr);
    fprintf('inner iterations started...\n')
    
    while ( size(param.Zscur, 1)<N+1)
        fprintf('sequential data size increment N: %d\n', size(param.Zscur, 1));
        np=np+1;
        % increase the R matrix by a row and column each time
        vec = [vec person(np)];
        Rin = param.R(vec,vec );
        settings.test_mask=test_mask(vec,vec);
        settings.train_mask=train_mask(vec,vec);
        
        if size(param.Zscur, 1)==1
            param.T;
        else
            param.T= [param.T zeros(1, size(param.T,1))'];
            param.T=[param.T; zeros(1, size(param.T,2))];
        end
        
        
        for k= 2:settings.in_itr
            [param, settings] = ...
                sampler_single_iter(Rin, param, settings);
        end
        
        % print some statistics during sequential initialisation
        numClusters=[];
        for i=1:length(param.NNscur)
            numClusters(i)=length(param.NNscur{i});
        end
        fprintf('\t M: %d avK: %d alpha: %d gamma: %d Bias: %d\n',size(param.Zscur,2), mean(numClusters), param.alpha, param.ggamma, param.bias);
        param.Zscur = [param.Zscur ; zeros(1, size(param.Zscur,2))];
        
    end
    
    param.Zscur(end,:) = [];
    disp('******* Sequential Initialization finished ******** ')
    
    % Rearrange the Z, C and I based on the real sequence of N objects and
    % not in the one provided by the 'person' vector.
    for ll=1:size(param.Zscur,2)
        a=param.Zscur(:,ll);
        a(find(a==1))=param.Cscur{:,ll};
        b(vec)= a;
        b(find(b==0))=[];
        param.Cscur{:,ll}=b;
        
        param.Zscur(vec, ll)=param.Zscur(:,ll);
        param.Iscur{ll}=find(param.Zscur(:,ll));
    end
    
    
    
end

% Store the result of the seuential initialisation as the first sample.
param.Zs_iter{1} = param.Zscur;
param.Is_iter{1} = param.Iscur;
param.Cs_iter{1} = param.Cscur;
param.Ks_iter{1} =param.Kscur;
param.Ns_iter{1}= param.Nscur;
param.NNs_iter{1}  = param.NNscur;
param.Ws_iter{1} =  param.Wscur;
param.alpha_iter(1)=param.alpha;
param.ggamma_iter(1)=param.ggamma;
param.bias_iter(1)=param.bias;


settings.test_mask=test_mask;
settings.train_mask=train_mask;

[param.lk  param.T ]=compute_likelihood(param.R, param.Zs_iter{1}, param.Cs_iter{1}, param.Is_iter{1}, param.Ws_iter{1}, param, settings);

param.E_iter{1} = compute_energy(param.R, param.Zs_iter{1}, param.Cs_iter{1}, param.Is_iter{1}, param.Ks_iter{1}, ...
    param.Ns_iter{1}, param.NNs_iter{1}, param.Ws_iter{1}, param, settings);


%% Start the actual sampler
for i = 2:settings.iterations
    fprintf('iteration: %d\n',i);
    
    [ param, settings] = ...
        sampler_single_iter(  param.R, param, settings);
    
    
    % --- store the samples
    param.Zs_iter{i}=param.Zscur;
    param.Cs_iter{i}=param.Cscur;
    param.Is_iter{i}=param.Iscur;
    param.Ks_iter{i}=param.Kscur;
    param.Ns_iter{i}=param.Nscur;
    param.NNs_iter{i}=param.NNscur;
    param.Ws_iter{i}=param.Wscur;
    param.ggamma_iter(i) = param.ggamma;
    param.alpha_iter(i) = param.alpha;
    param.bias_iter(i)= param.bias;
    
    
    %store the current energy
    param.E_iter{i} = compute_energy(param.R, param.Zs_iter{i}, param.Cs_iter{i}, param.Is_iter{i}, param.Ks_iter{i}, param.Ns_iter{i}, ...
        param.NNs_iter{i}, param.Ws_iter{i}, param, settings);
    
    % print some statistics
    fprintf('number of features: %d\n',size(param.Zs_iter{i},2));
    fprintf('          clusters:');
    for y=1:size(param.Zs_iter{i},2)
        fprintf('%d ',param.Ks_iter{i}(y) );
    end
    fprintf('\t alpha: %d gamma: %d Bias: %d\n', param.alpha, param.ggamma, param.bias);
    
    % plot the Z, C and W matrices
    if settings.plot_matrices==1
        hold on
        plot_ibp_matrices(param.Zs_iter{i}, param.Cs_iter{i}, param.Ws_iter{i} )
    end
    
    % Sanity check: make sure that the Z matrix does not contain all-zero
    % columns in the infinite case
    if ~isequal(settings.sampler_z, 'finite')
        for ch=1:size(param.Zs_iter{i},2)
            check_zero_col=sum(param.Zs_iter{i},1);
            if  find(check_zero_col==0)
                keyboard
            end
        end
    end
    toc
    
        
   %save(settings.results_file); 
    
end


